# Copyright 2020 Ecosoft Co., Ltd. (http://ecosoft.co.th)
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

from odoo import api, fields, models


class AccountMoveLine(models.Model):
    _name = "account.move.line"
    _inherit = ["account.move.line", "budget.docline.mixin"]
    _budget_date_commit_fields = ["move_id.date"]
    _budget_move_model = "account.budget.move"
    _doc_rel = "move_id"

    budget_move_ids = fields.One2many(
        comodel_name="account.budget.move",
        inverse_name="move_line_id",
        string="Account Budget Moves",
    )
    return_amount_commit = fields.Boolean(
        related="move_id.return_amount_commit",
    )

    @api.depends("move_id.not_affect_budget", "analytic_distribution")
    def _compute_can_commit(self):
        """
        Compute whether this account move line can commit budget.
        Checks if the move is marked as not affecting budget - if so,
        sets can_commit=False on those lines.
        """
        res = super()._compute_can_commit()
        no_budget_moves = self.mapped("move_id").filtered("not_affect_budget")
        if no_budget_moves:
            lines_to_update = no_budget_moves.mapped("line_ids").filtered("can_commit")
            if lines_to_update:
                lines_to_update.write({"can_commit": False})
        return res

    def recompute_budget_move(self):
        for invoice_line in self:
            invoice_line.budget_move_ids.unlink()
            # Commit on invoice
            invoice_line.commit_budget()

    def _init_docline_budget_vals(self, budget_vals, analytic_id):
        self.ensure_one()
        if self.move_id.move_type == "entry":
            total_amount = self.amount_currency
        else:
            sign = -1 if self.is_refund else 1
            discount = (100 - self.discount) / 100 if self.discount else 1
            total_amount = sign * self.price_unit * self.quantity * discount

        percent_analytic = sum(
            v / len(keys_list)
            for k, v in self[self._budget_analytic_field].items()
            if (keys_list := [x.strip() for x in k.split(",")])
            and str(analytic_id) in keys_list
        )

        # percent_analytic = self[self._budget_analytic_field].get(str(analytic_id))
        budget_vals["amount_currency"] = total_amount * (percent_analytic / 100)
        budget_vals["tax_ids"] = self.tax_ids.ids
        # Document specific vals
        budget_vals.update(
            {
                "move_line_id": self.id,
            }
        )
        return super()._init_docline_budget_vals(budget_vals, analytic_id)

    def _valid_commit_state(self):
        return self.move_id.state == "posted"

    def _get_included_tax(self):
        """Prepare for hook with extended modules"""
        if self._name == "account.move.line":
            return self.env.company.budget_include_tax_account
        return self.env["account.tax"]

    def uncommit_purchase_budget(self):
        """This function for hooks"""
        return
